/*
 * thm_studio.c
 *
 * Created: 21.08.2012 18:51:14
 *  Author: Nik
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

// 250ms frame @100Hz timer
#define FRAME_SIZE 25

#define SET_IN 0
#define THM_MAIN_IN 3
#define THM_AUX_IN 2

#define PIN_GLED 2
#define PIN_RLED 1
#define PIN_SCR 0

#define MAX_TEMP 390
    
volatile int frame=0;
volatile int setpoint =0;
volatile int oldsetpoint =0;
volatile int heater =0;
volatile uint8_t blinker=0;

volatile int tmpA = 0;
volatile int tmpB = 0;
//---------------------------------------
// in thermistors  ~= 6 pts/grad
//  limit to 10grad/sec @setpoint
#define FUZ_SPEED_LIMIT 60
// close to setpoint
#define FUZ_CLOSE 50
// 
#define HEAT_FULL 100
#define HEAT_ZERO 0
//---------------------------------------
int readADC(uint8_t ch)
{
ADMUX = ch & 0b0011;    

ADCSRA |= (1 << ADSC);  	
while ( ADCSRA & (1<<ADSC) ) ;

return ADC;
};
//---------------------------------------
inline void pinOn(uint8_t pin)
{ PORTB |= (1<<pin);}
//---------------------------------------
inline void pinOff(uint8_t pin)
{ PORTB &=  ( 0xFF ^ (1<<pin) ) ; }
//---------------------------------------
inline void offHeat(void) { pinOn(PIN_SCR); }
//---------------------------------------
inline void onHeat(void) { pinOff(PIN_SCR);	}
//---------------------------------------
void updSetpoint(void)
{
int sp = readADC(SET_IN);

oldsetpoint = setpoint;
setpoint = 397;
if( sp > 51 ) setpoint = 426;
if( sp > 102 ) setpoint = 456;
if( sp > 154 ) setpoint = 487;
if( sp > 205 ) setpoint = 519;
if( sp > 256 ) setpoint = 569;
if( sp > 307 ) setpoint = 618;
if( sp > 358 ) setpoint = 651;
if( sp > 510 ) setpoint = 0;

return;	
}
//---------------------------------------
ISR(TIM0_COMPA_vect) // main loop
{
frame++;
if(frame > FRAME_SIZE ) { frame = 0; } else { goto done; }

pinOn(PIN_GLED);
pinOff(PIN_RLED);

blinker++;

if(blinker & 1) {pinOff(PIN_GLED); }else{pinOn(PIN_GLED);}	

updSetpoint();
if( oldsetpoint != setpoint ) pinOn(PIN_RLED); //off red led for a frame
if( setpoint == 0 ) {heater = 0; offHeat();	goto done; }
// fuzzy processor
	
int oldtmpA = tmpA;	
int oldtmpB = tmpB;	

tmpA = readADC(THM_MAIN_IN);
tmpB = readADC(THM_AUX_IN);
int dtmpA = tmpA-oldtmpA;
int dtmpB = tmpB-oldtmpB;

int deltaA = setpoint-tmpA;
int deltaB = MAX_TEMP-tmpB;
// dtmp > 0 if cooling

int total = 0;
int power=0;

if( deltaA < 0 ) {total++; power+=HEAT_FULL; }
//if( deltaB < 0 ) {total++; power+=HEAT_FULL; }
if( deltaA >= 0 ) {total++; power+=HEAT_ZERO; }
if( deltaB >= 0 ) {total++; power+=HEAT_ZERO; }
// reduce speed 
if( (deltaA < 0 ) && ( dtmpA < -FUZ_SPEED_LIMIT ) ) {total++; power+=HEAT_ZERO; }
if( (deltaB < 0 ) && ( dtmpB < -FUZ_SPEED_LIMIT ) ) {total++; power+=HEAT_ZERO; }	
if( (deltaA < 0 ) && ( dtmpA > -FUZ_SPEED_LIMIT ) ) {total++; power+=HEAT_FULL; }
if( (deltaA < 0 ) && ( deltaA > -FUZ_CLOSE ) ) {total++; power+=HEAT_ZERO; }
if( (deltaA > 0 ) && ( dtmpA > 0 ) && ( deltaA < FUZ_CLOSE ) ) {total++; power+=HEAT_FULL; }		

// defuzzify 
if(total > 0 ){	power /= total; } else { power = 0; }
// heater control
heater += power;
if( heater >= HEAT_FULL ) { heater -= HEAT_FULL; onHeat(); pinOn(PIN_RLED);} else offHeat();
done:
power = 0;
}	
//---------------------------------------
int main(void)
{
	cli();
	
	DDRB = 0b000111; // 0 = SCR trigger , 1 - red led , 2 - green led
	// 5 - resistor, 3 - main t-res, 4 - aux t-res
	PORTB = 0b000011; // scr off, green on
	
	TCCR0A = (1<<WGM01); // WGM (010) = CTC
	TCCR0B = (1<<CS02)|(1<<CS00); // /1024
	
	OCR0A = 94; // yields 100HZ timer interrupt @ 9.6MHz
	TIMSK0 = (1<<OCIE0A);
	
	ADCSRA = (1 << ADPS2) | (1 << ADPS0); // /32  300k @ 9.6Mhz 
	ADCSRA |= (1 << ADEN);  // Enable ADC
	   
	readADC(0);
		
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_enable();
	sei(); // enable ints
	
    while(1)     { 	sleep_cpu();	}
}
//---------------------------------------